# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/3/21 14:55
Desc:
'''
import pandas as pd
import numpy as np
import dill

class ECardStandardized(object):
    def __init__(self):
        self.clf = None
        self.score_card = None
        self.score_bins = None
        self.card_ext = pd.DataFrame()

    def load(self, model_name:str):
        model = dill.load(open(model_name, "rb"))
        self.clf = model.get('model')
        self.score_card  = model.get('score_card')
        self.score_bins = model.get('score_bins',[])
        self.card_ext = model.get('features_ext',pd.DataFrame())
        self.info = model.get('info','')

        self.preprocessing_befor = model.get('preprocessing_befor',self.func_empty)
        self.preprocessing_after = model.get('preprocessing_after',self.func_empty)
        self.postprocessing = model.get('postprocessing',self.get_g7_level)

    def dump(self, clf_model, model_name, model='all'):
        assert model in ['all','pure'],'all for backup, pure for online'
        score_card = clf_model.score_ecard
        card_ext = clf_model.feature_ext_info.get("df_ext",pd.DataFrame())
        assert hasattr(clf_model,'score_bins'), "need to set 'clf_model.score_bins=[]' for score_level"
        score_bins = clf_model.score_bins
        if hasattr(clf_model, 'info'):
            info = clf_model.info
        else:
            info = ""
        if model =='pure':
            model_raw = None
        else:
            model_raw = clf_model
        model = {
            'model': model_raw,
            'score_card': score_card,
            'features_ext': card_ext,
            'score_bins': score_bins,
            'info': info
        }
        dill.dump(model, open(model_name, "wb"), protocol=3)

    def predict(self,data):
        score_info = self.get_single_score(data)
        return score_info

    def get_single_score(self, data=None):
        if data is None:
            data = {}
        data = self.preprocessing_befor(data)
        data = self.get_ext_features(data)
        data = self.preprocessing_after(data)
        score = 0
        score_detail = {}
        for i, row in self.score_card.iterrows():
            field_ = row['field_']
            bins_ = row['bins_']
            score_ = row['score_']
            if field_ in ['init_base_score', 'init_model_score']:
                v = 0
            else:
                v = data.get(field_)
            try:
                v = float(v)
            except:
                score = 0
                score_detail[field_] = '异常'
                break
            if not (v > -np.inf):
                v = 0
            if v in bins_:
                score = score + score_
                score_detail[field_] = score_
            else:
                pass
        score = round(score,2)
        level = np.argwhere(np.sort(self.score_bins + [score]) == score).min()

        # level
        score,level = self.postprocessing(data,score,level)
        out = {'card_score': score, 'card_level': level, 'score_detail': score_detail}
        return out

    def get_ext_features(self, raw_data):
        data = raw_data.copy()
        df_ext = self.card_ext.copy()
        ext_columns = [col_ for col_ in df_ext.columns if 'ext-' in col_]
        for icol in ext_columns:
            colinfo = str(icol).split('-')
            if 'ext' == colinfo[0]:
                idx_equ = True
                for jcol in colinfo[1:]:
                    if jcol not in df_ext.columns:
                        break
                    idx_equ = idx_equ & df_ext[jcol].apply(lambda _: data.get(jcol) in _).astype(bool)
                value_ = df_ext.loc[idx_equ, icol].values
                if len(value_) > 0:
                    data.update({icol: value_[0]})
                else:
                    data.update({icol: 0})
        return data

    def func_empty(self,x):
        return x

    def get_g7_level(self,data, score, level):
        total_meters = data.get('run_meters', 0)
        total_seconds = data.get('run_seconds', 0)
        if (total_meters < 2000000) or (total_seconds < 10 * 3600):
            level = -1
        if score<=0:
            level = -1
        return score,level

if __name__ == '__main__':
    model_name = "../data/score_card_model_online_o2o_tangshan_v3.2.0.pkl"
    ecard = ECardStandardized()
    ecard.load(model_name)

    data = {
        'carnum': 'LBZ447DB1KA009342',
        'trip_cnt': 143,
        'run_meters': 12275107.380000005,
        'run_seconds': 771092,
        'trip_avg_meters': 85839.91174825176,
        'trip_avg_seconds': 5392.251748251748,
        'trip_avg_distance': 44253.81818181818,
        'high_meters_ratio': 0.002143811796129477,
        'province_meters_ratio': 0.34486200315422405,
        'high_trip_cnt_ratio': 0.0,
        'province_trip_cnt_ratio': 0.5874125874125874,
        'curvature_g2_trip_meters_ratio': 0.5600617817161611,
        'ng_23_6_seconds_ratio': 0.2924709373200604,
        'ng_23_6_trip_cnt_ratio': 0.4335664335664336,
        'daily_run_kmeters': 84.41548247018672,
        'daily_run_hours': 1.4729923918800238,
        'daily_trip_cnt': 0.9834059792344316,
        'daily_nohigh_kmeters': 84.23451156309116,
        'daily_ng_23_6_hours': 0.43080746551846827,
        'trip_long_cnt_ratio': 0.0,
        'day_high_meters_ratio': 0.0017513956769965135,
        'ng_province_meters_ratio': 0.10545102620519804,
        'morn_6_10_seconds_ratio': 0.13508245449310846,
        'dusk_17_20_seconds_ratio': 0.19514792009254409,
        'ng_23_6_avg_speed': 60.85796919147581,
        'morn_6_10_avg_speed': 55.72396178992139,
        'dusk_17_20_avg_speed': 51.151944177515524,
        'low_speed_seconds_ratio': 0.3662662821038216,
        'low_speed_block_cnt_ratio': 0.2351517298382336,
        'week_1_5_seconds_ratio': 0.29913551171585234,
        'geohash4_top10_meters_ratio': 0.9702472802319388,
        'trip_r30m_cnt_ratio': 0.4795918367346938,
        'common_line_top30_cnt_ratio': 0.030769230769230767,
        'mil_ratio_province_hb': 1.0,
        'mil_ratio_province_tj': 0.0,
        'mil_ratio_province_else': 0.0,
        'mil_ratio_city_ts': 1.0,
        'mil_ratio_city_around': 0.0,
        'mil_ratio_city_else': 0.0,
        'mil_ratio_county_west': 0.0015470292366599212,
        'mil_ratio_county_south': 0.5677351500284784,
        'mil_ratio_county_east': 0.2494958703978352,
        'mil_ratio_county_else': 0.18122195033702654,
        'top1_city_mileage_rate': 1.0,
        'top2_city_mileage_rate': 1.0,
        'top1_province_mileage_rate': 1.0,
        'top2_province_mileage_rate': 1.0,
        'top1_county_mileage_rate': 0.3229254357854749,
        'top2_county_mileage_rate': 0.5677351500284797,
        'ratio_nonglinmufu': 0.0290325496489345,
        'ratio_meitan': 0.0006003401293231343,
        'ratio_gangtie': 0.1395834542570436,
        'ratio_shashi': 0.017055577543834046,
        'ratio_kuaidi': 0.08425476464525228,
        'ratio_jiadian': 0.004164982562417855,
        'ratio_fengdian': 0.0,
        'ratio_jixie': 0.6604690662393785,
        'ratio_qiche': 0.0006478740548063132,
        'ratio_other': 0.06419139091900974,
        'pred': '机械',
        'triggertime': '2021-04-02 02:35:52',
        'triggertime_end': '2021-08-29 22:33:06',
        'source': 10
     }
    out = ecard.predict(data)
    print(out)


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 22:08
Desc:
'''

import numpy as np
import pandas as pd
from xgboost import XGBClassifier
from sklearn.metrics import classification_report
from score_ecard.features import layered_woe as woe
from score_ecard.features import xgboost_blocks as xgb_bolcks
from score_ecard.util import progress_bar, log_step

class XGBoostWoe:
    def __init__(self, xgb_params):
        self.xgb_params = xgb_params
        self.eps = 1e-4
        self.woe_cards = []

    def fit(self,df_X,df_Y,sample_weight=None):
        '''

        :param df_X: 特征数据
        :param df_Y: 标签分类字段
        :param sample_weight:
        :return: 返回woe详细信息
        '''
        log_step("XGBoostWoe: (start) ->->->  ")
        clf_xgb = XGBClassifier(**self.xgb_params)
        clf_xgb.fit(df_X, df_Y, sample_weight=sample_weight)
        xgb_report = classification_report(df_Y, clf_xgb.predict(df_X))
        log_step("XGB-report: \n {}".format(xgb_report))

        xgb_boundaries = xgb_bolcks.get_xgboost_blocks(clf_xgb, col_name=df_X.columns)
        log_step("Total {} trees.".format(len(xgb_boundaries)))

        if len(clf_xgb.classes_)>2:
            classes_list = clf_xgb.classes_.tolist() * clf_xgb.n_estimators
        else:
            classes_list = clf_xgb.classes_.tolist()[:1] * clf_xgb.n_estimators
        assert len(classes_list)==len(xgb_boundaries)

        data = []
        for i, tree_bins in enumerate(xgb_boundaries):
            progress_bar(i,len(xgb_boundaries), is_pass=False)
            y_ = (df_Y == classes_list[i]).astype(int)
            df_woe, x = woe.get_woe_card(df_X, y_, tree_bins)
            self.woe_cards.append(df_woe)
            data.append(x)
        log_step("\nXGBoostWoe: <-<-<- (end). ")

        return data

    def transform(self, df_X, num_trees=None):
        '''

        :param df_X:
        :param num_trees: 需要转义的第i棵树，默认None表示全部的树进行woe转换
        :return:
        '''
        if num_trees is None:
            data=[]
            for woe_card in self.woe_cards:
                df_data = pd.DataFrame()
                for field_ in woe_card.field_.unique():
                    df_woe = woe_card[woe_card.field_ == field_]
                    bins_ = np.sort(list(set(df_woe.boundary_.unique().tolist() + [np.inf, -np.inf]))).tolist()
                    x_bins = pd.cut(df_X[field_], bins=bins_)
                    df_bins = pd.DataFrame(x_bins)
                    df_bins.columns = ['bins_']
                    out = pd.merge(df_bins, df_woe.set_index('bins_'), how='left', left_on='bins_', right_index=True)
                    df_data[field_+'_woe1'] = out['woe1']
                data.append(df_data)
            return data
        else:
            assert -len(self.woe_cards) <= num_trees<len(self.woe_cards)
            woe_card = self.woe_cards[num_trees]
            df_data = pd.DataFrame()
            for field_ in woe_card.field_.unique():
                df_woe = woe_card[woe_card.field_ == field_]
                bins_ = np.sort(list(set(df_woe.boundary_.unique().tolist() + [np.inf, -np.inf]))).tolist()
                x_bins = pd.cut(df_X[field_], bins=bins_)
                df_bins = pd.DataFrame(x_bins)
                df_bins.columns = ['bins_']
                out = pd.merge(df_bins, df_woe.set_index('bins_'), how='left', left_on='bins_', right_index=True)
                df_data[field_+'_woe1'] = out['woe1']

            return df_data


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 12:23
Desc:
'''

import numpy as np

def get_xgboost_blocks(clf_xgb, col_name=None, valid_feature_prob=1.0):
    '''
    利用xgb获得最优分箱的边界值列表
    :param clf_xgb:
    :param col_name:
    :param valid_feature_prob: 选择随机森林有效特征的比例，剔除低重要度特征，建议在0.85~1.0
    :return:
    '''
    importances = clf_xgb.feature_importances_
    indices = np.argsort(importances)[::-1]
    col_set = col_name[importances[indices].cumsum() <= valid_feature_prob]
    booster = clf_xgb.get_booster()
    df_trees = booster.trees_to_dataframe()
    df_trees = df_trees[df_trees.Feature.isin(col_set)]
    df_boundary = df_trees.groupby(['Tree', 'Feature'])['Split'].apply(
        lambda x: sorted([-np.inf,np.inf]+list(set(x)))).unstack().applymap(
        lambda x: None if str(x)=='nan' else x)
    xgb_boundary = [dict(j.dropna()) for i, j in df_boundary.iterrows()]
    return xgb_boundary



# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/4/20 11:31
Desc:
'''
from xgboost import XGBClassifier
import pandas as pd

df_valid = pd.read_csv("../score_ecard/data/train_test_data.csv")
df_train_data = df_valid[df_valid['train_test_tag'] == '训练集'].fillna(0).head(10000)
df_test_data = df_valid[df_valid['train_test_tag'] == '测试集'].fillna(0).head(10000)
feature_columns = df_train_data.columns[4:33].tolist()
feature_columns.extend(df_train_data.columns[36:46])
df_X = df_train_data[feature_columns]
# df_Y = df_train_data['label']
df_Y=df_train_data.apply(lambda x:x['label'] if x["report_fee"]<5000 else 2,axis=1)

params_xgb = {
            'n_estimators': 3,
            'max_depth':3,
            'eta': 0.3,
            'min_child_weight': 1,
            'gamma': 0.3,
            'subsample': 0.8,
            'colsample_bytree': 0.8,
            'booster': 'gbtree',
            'objective': 'binary:logistic',
            'scale_pos_weight': 1,
            'lambda': 1,
            'seed':666,
            'silent': 0,
            'eval_metric': 'auc'
        }
clf_xgb = XGBClassifier(**params_xgb)
clf_xgb.fit(df_X, df_Y)
aa=clf_xgb.get_booster()
aa.trees_to_dataframe()



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-25 14:40
Desc:
'''

from .util import *
from .ecard_explanation import summary_explanation,factor_trend,model_factor_report
# from .ecard_explanation import sample_explanation
from ..features.factor_screening import factor_screening


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/3/31 11:49
Desc:
'''
import pandas as pd

def factor_screening(df_x:pd.DataFrame, df_label:pd.Series, df_report_fee:pd.Series=None, frac_label_stability=0.2,
                    frac_report_stability=0.1,sampling_frequency=5):
    '''
    ## 1.特征筛选，抽样n次中划分的两部分数据分布均区域一致的topn因子，即稳定的因子
    :param df_x: 特征集合
    :param df_label: 是否出险 或 出险次数
    :param df_report_fee: 赔付金额
    :return: 风险表征能力稳定的特征
    :param frac_label_stability: 因子筛选出险率占比
    :param frac_report_stability: 因子筛选赔付占比
    :param sampling_frequency: 抽样频次
    '''
    if df_report_fee is None:
        df_report_fee = df_label
    assert df_x.shape[0] == df_label.shape[0]==df_report_fee.shape[0]>2000
    df_x = pd.DataFrame(df_x)
    df_label = pd.DataFrame(df_label)
    df_label.columns=['report_num']
    df_report_fee = pd.DataFrame(df_report_fee)
    df_report_fee.columns=['report_fee']
    df_data = pd.concat([df_x,df_label,df_report_fee],axis=1)

    df_indicators_report = pd.DataFrame()
    for i in range(sampling_frequency):
        df_stability = df_data.copy()
        group_idx = df_stability.sample(frac=0.5, random_state=666+i).index
        df_stability['group_'] = df_stability.index.isin(group_idx).astype(int)

        for icol in df_x.columns:
            if df_stability[icol].unique().size == 1:
                continue
            df_stability[icol] = pd.qcut(df_stability[icol], q=10, duplicates='drop')
            df_indactors_tmp = df_stability.groupby([icol, 'group_']).agg(
                {
                    'report_num': [("出险次数", 'sum')],
                    'report_fee': [("赔付金额", 'sum')],
                }
            ).droplevel(0, 1).unstack().sort_index()
            df_indicators_report.loc[icol,"出险次数_t{}".format(i)] = df_indactors_tmp.出险次数.corr(method='pearson')[0][1]
            df_indicators_report.loc[icol,"赔付金额_t{}".format(i)] = df_indactors_tmp.赔付金额.corr(method='kendall')[0][1]
    chuxian_col = [c for c in df_indicators_report.columns if "出险次数" in c]
    peifu_col = [c for c in df_indicators_report.columns if "赔付金额" in c]
    df_indicators_report[chuxian_col] = df_indicators_report[chuxian_col].apply(
        lambda x: x > x.quantile(frac_label_stability), axis=0)
    df_indicators_report[peifu_col] = df_indicators_report[peifu_col].apply(
        lambda x: x > x.quantile(frac_report_stability), axis=0)
    valid_factor_list = list(df_indicators_report[df_indicators_report.all(axis=1)].index)

    return valid_factor_list


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@163.com
Date: 2021-11-5 15:17
Desc:
'''

import sys
import math
import inspect
from functools import wraps
import numpy as np
import time
import logging

# 1.进度条展示 progress_bar(1, 100)
def progress_bar(portion, total, is_pass=True):
    """
    total 总数据大小，portion 已经传送的数据大小
    :param portion: 已经接收的数据量
    :param total: 总数据量
    :return: 接收数据完成，返回True
    """
    if is_pass:
        return True
    part = total / 50  # 1%数据的大小
    count = math.ceil(portion / part)
    sys.stdout.write('\r')
    sys.stdout.write(('[%-50s]%.2f%%' % (('>' * count), portion / total * 100)))
    sys.stdout.flush()

    if portion >= total:
        sys.stdout.write('\n')
        return True
    return False

# 2.获取变量名
def get_retrieve_name(var):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    list_name = [var_name for var_name, var_val in callers_local_vars if var_val is var]
    return list_name

# 3.计算加权方差
def get_weighted_std(values,weights):
    average = np.average(values, weights=weights)
    variance = np.average((values-average)**2, weights=weights)
    return variance

# 4.计算等频分箱
def get_frequence_blocks(x, box_num=10, drop_duplicates=True):
    '''
        等频分箱
    :param x: dtype=[]
    :param bins: 分箱数量
    :return: 箱边界
    '''
    x=np.sort(x).astype(float)
    len_clocks = int(min(len(x),box_num))
    ind = np.linspace(0, len(x)-1, len_clocks+1).astype(int)
    blocks=np.array(x)[ind].astype(float)
    if drop_duplicates:
        blocks = np.unique(blocks)
    if len(blocks)<2:
        return [-np.inf,np.inf]
    blocks[0] = -np.inf
    blocks[-1] = np.inf
    return blocks.tolist()

# 4.计算等距分箱
def get_distince_blocks(x, box_num=5, drop_duplicates=True):
    '''
        等距分箱
    :param x: dtype=[]
    :param bins: 分箱数量
    :return: 箱边界
    '''
    blocks = np.linspace(min(x), max(x), box_num+1).astype(float)
    if drop_duplicates:
        blocks = np.unique(blocks)
    if len(blocks)<2:
        return [-np.inf,np.inf]
    blocks[0] = -np.inf
    blocks[-1] = np.inf
    return blocks.tolist()

# 5.函数执行时间
def log_run_time(func):
    logging.basicConfig(level='INFO')
    logger = logging.getLogger()
    @wraps(func)
    def wrapper(*args, **kwargs):
        func_name = func.__name__
        if kwargs.get('log', True):
            func_start_time = time.time()
            logger.info('{} Start {}'.format(func_name, time_format(func_start_time)))
        result = func(*args, **kwargs)
        if kwargs.get('log', True):
            func_end_time = time.time()
            func_runtime = func_end_time - func_start_time
            logger.info('{} End {}, Runtime {} S'.format(func_name, time_format(func_end_time), func_runtime))
        return result
    return wrapper

def log_cur_time(info=""):
    logging.basicConfig(level='INFO')
    logger = logging.getLogger()
    lineno = sys._getframe().f_back.f_lineno
    co_name = sys._getframe().f_back.f_code.co_name
    logger.info('{}:lineno_{} Start {}{}'.format(co_name,lineno, time_format(time.time()), info))

def log_step(info):
    logging.basicConfig(level='INFO')
    logger = logging.getLogger()
    logger.info('{} {}'.format(time_format(time.time()),info))

def time_format(time_value):
    local_time = time.localtime(time_value)
    data_head = time.strftime("%Y-%m-%d %H:%M:%S", local_time)
    data_secs = (time_value - int(time_value)) * 1000
    time_stamp = "%s.%03d" % (data_head, data_secs)
    return time_stamp


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 12:23
Desc:
'''

import numpy as np
from collections import OrderedDict

def get_randomforest_blocks(clf_rf, col_name=None, valid_feature_prob=1.0, cross_hierarchy=0):
    '''
    利用随机森林获得最优分箱的边界值列表
    :param clf_rf:
    :param col_name:
    :param valid_feature_prob: 选择随机森林有效特征的比例，剔除低重要度特征，建议在0.85~1.0
    :param cross_hierarchy: 各决策树内进行特征交叉的层深
    :return:
    rf_cross_boundary: 各决策树的前n层特征及boundary，用来特征交叉
    '''
    rf_boundary = []
    rf_cross_boundary = []
    importances = clf_rf.feature_importances_
    indices = np.argsort(importances)[::-1]
    col_set = indices[importances[indices].cumsum() <= valid_feature_prob]
    for i, treei in enumerate(clf_rf.estimators_):
        n_nodes = treei.tree_.node_count
        feature_idx = treei.tree_.feature
        if col_name is not None:
            feature = np.array([(col_name[i] if i>=0 else i)for i in feature_idx])
        else:
            feature = feature_idx
        threshold = treei.tree_.threshold
        children_left = treei.tree_.children_left
        children_right = treei.tree_.children_right
        tree_boundary = {}
        for j in range(n_nodes):
            if feature_idx[j] not in col_set:
                continue
            if children_left[j] != children_right[j]:  # 获得决策树节点上的划分边界值(除叶子结点)
                if feature[j] in tree_boundary:
                    boundary = tree_boundary.get(feature[j])
                    boundary.append(threshold[j])
                else:
                    boundary = [-np.inf, np.inf]
                    boundary.append(threshold[j])
                tree_boundary.update({feature[j]: boundary})
        for k, v in tree_boundary.items():
            sv = list(set([round(i, 4) for i in v]))
            sv.sort()
            tree_boundary.update({k: sv})
        rf_boundary.append(tree_boundary)

        # extend cross feature
        ext_features = OrderedDict()
        if cross_hierarchy >= 2:
            step_n = cross_hierarchy
            cross_feature = {feature[0]: [-np.inf, threshold[0], np.inf]}
            cur_list = [0]
            while (step_n > 1):
                next_list = []
                for cur_ in cur_list:
                    left_ = children_left[cur_]
                    right_ = children_right[cur_]
                    for i in [left_, right_]:
                        if i < 0:
                            continue
                        cur_feature = feature[i]
                        cur_threshold = threshold[i]
                        if cur_feature in cross_feature:
                            boundary_ = cross_feature.get(cur_feature)
                            boundary_.append(cur_threshold)
                        else:
                            boundary_ = [-np.inf, cur_threshold, np.inf]
                        cross_feature.update({cur_feature: boundary_})
                        next_list.append(i)
                cur_list = next_list
                step_n -= 1
            col_sorted = col_name[indices] if col_name is not None else indices
            for icol in col_sorted:
                v = cross_feature.get(icol)
                if v:
                    sv = list(set([round(i, 4) for i in v]))
                    sv.sort()
                    ext_features.update({icol: sv})
        rf_cross_boundary.append(ext_features)
    return rf_boundary, rf_cross_boundary






# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 21:51
Desc:
'''
import numpy as np
import pandas as pd
import random
import itertools
import warnings
warnings.filterwarnings('ignore')

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier
from multiprocessing import Process, Manager

from score_ecard import score_card
from score_ecard.features import layered_woe as woe
from score_ecard.features import randomforest_blocks as rf_bolcks
from score_ecard.features import xgboost_blocks as xgb_bolcks
from score_ecard.util import progress_bar, get_weighted_std, log_step,summary_explanation,model_factor_report
from score_ecard.util.ecard_standardized import ECardStandardized

class ECardModel(object):
    def __init__(self,kwargs_lr=None, features_model='rf', kwargs_rf=None, kwargs_xgb=None, sample_frac=1.0,
                 cross_hierarchy=0, is_best_bagging=False, optimize_type='global', n_estimators=None):
        '''

        :param kwargs_lr:
        :param features_model:
        :param kwargs_rf:
        :param kwargs_xgb:
        :param sample_frac: 样本再抽样比例，样本抽样比例 = 0.6*sample_frac
        :param cross_hierarchy: 特征交叉层级，可参考参数设置为[2，3]，默认不进行特征扩展即cross_hierarchy=0
        :param is_best_bagging:
        :param optimize_type: 优化方式={'local': 局部最优, 'global': 全局最优}
        '''
        self.params_lr = {
            "class_weight": 'balanced',
            "penalty": 'l1',
            "C": 0.2,
            "solver": 'liblinear'
        }
        self.params_rf = {
            'n_estimators': 100,
            'max_depth': 5,
            'max_features': 'auto',
            'min_samples_leaf': 0.01,
            'bootstrap': True,
            'class_weight': "balanced",
            'random_state': 666
        }
        self.params_xgb = {
            'n_estimators': 10,
            'max_depth':5,
            'eta': 0.3,
            'min_child_weight': 1,
            'gamma': 0.3,
            'subsample': 0.8,
            'colsample_bytree': 0.8,
            'booster': 'gbtree',
            'objective': 'binary:logistic',
            'scale_pos_weight': 1,
            'lambda': 1,
            'seed':666,
            'silent': 0,
            'eval_metric': 'auc'
        }
        self.params_base_tree = {
            'max_depth': 3,
            'max_features': None,
            'min_samples_leaf': 0.01,
            'class_weight': "balanced",
            'random_state': 666
        }
        if kwargs_lr:
            self.params_lr.update(**kwargs_lr)
        if kwargs_rf:
            self.params_rf.update(**kwargs_rf)
        if kwargs_xgb:
            self.params_xgb.update(**kwargs_xgb)
        self.features_model = features_model
        self.sample_frac = sample_frac
        self.cross_hierarchy = cross_hierarchy
        self.is_best_bagging = is_best_bagging
        self.optimize_type = optimize_type
        self.init_cards = []
        self.score_ecard = None
        self.cur_card_index = 0
        self.feature_ext_info = {}
        self.score_model = score_card.ScoreCardModel()
        random.seed(666)
        if n_estimators is None:
            self.n_estimators = self.params_rf.get("n_estimators",
                                                   100) if features_model == 'rf' else self.params_xgb.get("n_estimators",
                                                                                                           100)
        else:
            self.n_estimators = n_estimators

    def fit(self, df_X, df_Y, validation_X:pd.DataFrame=None, validation_Y:pd.DataFrame=None, sample_weight=None, validation_step=1):
        self.init_check()
        assert df_X.shape[0] == df_Y.shape[0]
        df_Y, df_y, df_y_binary= self.check_label(df_Y)
        if type(sample_weight)==type(None):
            sample_weight=df_y*0+1

        log_step("start model fitting ...")
        df_X = self.calc_base_card(df_X, df_Y, df_y, df_y_binary, sample_weight)
        ml_boundaries = self.get_ml_boundaries(df_X, df_y, sample_weight=sample_weight)
        random.shuffle(ml_boundaries)
        init_card, df_X_bins = self.get_init_ecard(df_X, ml_boundaries)

        validation_idx = False
        if type(validation_X) != type(None):
            assert validation_X.shape[0] == validation_Y.shape[0]
            df_valY, df_valy, df_valy_binary = self.check_label(validation_Y)
            validation_X = self.get_cross_features(validation_X)
            df_valX_bins = pd.DataFrame()
            for col_, bins_ in self.blocks.items():
                idata = pd.cut(validation_X.loc[:, col_].fillna(0), bins=bins_)
                df_valX_bins[col_] = idata
            validation_idx = True

        log_step("start model iterative optimization ...")
        best_auc = -999
        best_insurance_auc = -999
        best_card = init_card
        for i,tree_bins in enumerate(ml_boundaries):
            if i>self.n_estimators:
                continue
            sample_idx = df_X.sample(
                frac=1.0, replace=True, weights=sample_weight, random_state=i
            ).sample(frac=self.sample_frac, replace=False, random_state=i).index
            df_woe,x = woe.get_woe_card(df_X.loc[sample_idx], df_Y.loc[sample_idx], tree_bins)
            clf_lr = LogisticRegression(**self.params_lr)
            clf_lr.fit(x, df_y_binary.loc[sample_idx], sample_weight=sample_weight.loc[sample_idx])
            clf_lr.col_name = x.columns
            add_card = self.get_score_card(clf_lr, df_woe)
            cur_card = self.get_cards_merge(best_card, add_card,
                                            cur_weight=self.cur_card_index / (self.cur_card_index + 1),
                                            add_weight=1.0 / (self.cur_card_index + 1))
            if 'local'==self.optimize_type:
                idx_ = sample_idx
            else:
                idx_ = df_X_bins.index
            df_pre_score = self.calc_score(cur_card, df_X_bins.loc[idx_], is_woe_feature = True)
            predict_ = -1*df_pre_score
            train_auc = self.score_model.get_auc(predict_, df_y_binary.loc[idx_].copy(), pre_target=1)[0]
            train_info = "train_auc={}".format(round(train_auc, 4))
            if self.is_best_bagging and (train_auc < best_auc):
                vote = False
            else:
                best_auc = train_auc
                vote = True
            if ('fee_got' in df_Y.columns) and ('report_fee' in df_Y.columns):
                train_insurance_auc = \
                    self.score_model.get_insurance_auc(predict_, df_Y.loc[idx_]["fee_got"], df_Y.loc[idx_]["report_fee"])[0]
                train_info = "{}, train_insurance_auc={}".format(train_info, round(train_insurance_auc, 4))
                if train_insurance_auc >= best_insurance_auc:
                    best_insurance_auc = train_insurance_auc
                    vote = True
            if vote:
                self.cur_card_index += 1
                best_card = cur_card

            validation_info=None
            if validation_idx and vote and (i%validation_step==0):
                df_pre_score = self.calc_score(cur_card, df_valX_bins, is_woe_feature=True)
                predict_ = -1 * df_pre_score
                validation_auc = self.score_model.get_auc(predict_, df_valy_binary.copy(), pre_target=1)[0]
                validation_info = "validation_auc={}".format(round(validation_auc, 4))
                if ('fee_got' in df_valY.columns) and ('report_fee' in df_valY.columns):
                    validation_insurance_auc = self.score_model.get_insurance_auc(predict_, df_valY["fee_got"], df_valY["report_fee"], )[0]
                    validation_info = "{}, validation_insurance_auc={}".format(validation_info,round(validation_insurance_auc,4))
            log_step("step_{}:\t{}\t{}".format(i+1,train_info,validation_info))
        self.score_ecard = best_card
        self.score_lower = best_card.groupby("field_").score_.min().sum()
        self.score_upper = best_card.groupby("field_").score_.max().sum()

    def fit_parallar(self, df_X, df_Y, validation_X:pd.DataFrame=None, validation_Y:pd.DataFrame=None, sample_weight=None, validation_step=1):
        assert df_X.shape[0] == df_Y.shape[0]
        df_Y, df_y, df_y_binary= self.check_label(df_Y)
        if type(sample_weight)==type(None):
            sample_weight=df_y*0+1

        log_step("start model fitting ...")
        df_X = self.calc_base_card(df_X, df_Y, df_y, df_y_binary, sample_weight)
        ml_boundaries = self.get_ml_boundaries(df_X, df_y, sample_weight=sample_weight)
        random.shuffle(ml_boundaries)
        init_card, df_X_bins = self.get_init_ecard(df_X, ml_boundaries)

        validation_idx = False
        if type(validation_X) != type(None):
            assert validation_X.shape[0] == validation_Y.shape[0]
            df_valY, df_valy, df_valy_binary = self.check_label(validation_Y)
            validation_X = self.get_cross_features(validation_X)
            df_valX_bins = pd.DataFrame()
            for col_, bins_ in self.blocks.items():
                idata = pd.cut(validation_X.loc[:, col_].fillna(0), bins=bins_)
                df_valX_bins[col_] = idata
            validation_idx = True

        log_step("start model iterative optimization ...")
        def fit_card(i,x_,y_,y_binary_,weight_,tree_bins_,return_dict):
            df_woe, x = woe.get_woe_card(x_, y_, tree_bins_)
            clf_lr = LogisticRegression(**self.params_lr)
            clf_lr.fit(x, y_binary_, weight_)
            clf_lr.col_name = x.columns
            add_card = self.get_score_card(clf_lr, df_woe)
            return_dict[i] = add_card

        return_dict = Manager().dict()
        jobs = []
        for i, tree_bins_ in enumerate(ml_boundaries):
            if i > self.n_estimators:
                continue
            sample_idx = df_X.sample(
                frac=1.0, replace=True, weights=sample_weight, random_state=i
            ).sample(frac=self.sample_frac, replace=False, random_state=i).index
            x_ = df_X.loc[sample_idx]
            y_ = df_Y.loc[sample_idx]
            y_binary_ = df_y_binary.loc[sample_idx]
            weight_ = sample_weight.loc[sample_idx]
            p = Process(target=fit_card, args=(i,x_,y_,y_binary_,weight_,tree_bins_,return_dict))
            jobs.append(p)
            p.start()
        for proc in jobs:
            proc.join()

        best_auc = -999
        best_insurance_auc = -999
        best_card = init_card
        for add_card in return_dict.values():
            cur_card = self.get_cards_merge(best_card, add_card,
                                            cur_weight=self.cur_card_index / (self.cur_card_index + 1),
                                            add_weight=1.0 / (self.cur_card_index + 1))
            if 'local'==self.optimize_type:
                idx_ = sample_idx
            else:
                idx_ = df_X_bins.index
            df_pre_score = self.calc_score(cur_card, df_X_bins.loc[idx_], is_woe_feature = True)
            predict_ = -1*df_pre_score
            train_auc = self.score_model.get_auc(predict_, df_y_binary.loc[idx_].copy(), pre_target=1)[0]
            train_info = "train_auc={}".format(round(train_auc, 4))
            if self.is_best_bagging and (train_auc < best_auc):
                vote = False
            else:
                best_auc = train_auc
                vote = True
            if ('fee_got' in df_Y.columns) and ('report_fee' in df_Y.columns):
                train_insurance_auc = \
                    self.score_model.get_insurance_auc(predict_, df_Y.loc[idx_]["fee_got"], df_Y.loc[idx_]["report_fee"])[0]
                train_info = "{}, train_insurance_auc={}".format(train_info, round(train_insurance_auc, 4))
                if train_insurance_auc >= best_insurance_auc:
                    best_insurance_auc = train_insurance_auc
                    vote = True
            if vote:
                self.cur_card_index += 1
                best_card = cur_card

            validation_info=None
            if validation_idx and vote and (i%validation_step==0):
                df_pre_score = self.calc_score(cur_card, df_valX_bins, is_woe_feature=True)
                predict_ = -1 * df_pre_score
                validation_auc = self.score_model.get_auc(predict_, df_valy_binary.copy(), pre_target=1)[0]
                validation_info = "validation_auc={}".format(round(validation_auc, 4))
                if ('fee_got' in df_valY.columns) and ('report_fee' in df_valY.columns):
                    validation_insurance_auc = self.score_model.get_insurance_auc(predict_, df_valY["fee_got"], df_valY["report_fee"], )[0]
                    validation_info = "{}, validation_insurance_auc={}".format(validation_info,round(validation_insurance_auc,4))
            log_step("step_{}:\t{}\t{}".format(i+1,train_info,validation_info))
        self.score_ecard = best_card
        self.score_lower = best_card.groupby("field_").score_.min()
        self.score_upper = best_card.groupby("field_").score_.max()

    def check_label(self, df_label):
        df_Y=df_label.copy()
        if df_Y.shape[-1] == 1:
            df_Y.columns = ['label']
        if 'label' not in df_Y.columns:
            df_Y['label'] = df_Y.iloc[:,0]
        df_y=df_Y.label.astype(int)
        if df_y.unique().size>2:
            df_y_binary = df_y.apply(lambda x: 1 if x>0 else 0)
        else:
            df_y_binary = df_y
        df_Y['label']=df_y_binary
        ## test---
        if 'label_binary' in df_Y.columns:
            df_y_binary = df_Y["label_binary"]
        ## ---end
        return df_Y, df_y, df_y_binary

    def calc_base_card(self,df_X, df_Y, df_y, df_y_binary, sample_weight):
        clf_base_tree = DecisionTreeClassifier(**self.params_base_tree)
        clf_base_tree.fit(df_X, df_y, sample_weight=sample_weight)
        if hasattr(clf_base_tree, 'estimators_'):
            pass
        else:
            clf_base_tree.estimators_ = [clf_base_tree]
        tree_boundaries, cross_fields_boundaries = rf_bolcks.get_randomforest_blocks(clf_base_tree, col_name=df_X.columns, cross_hierarchy=self.cross_hierarchy)
        tree_bins, cross_fields_bins = tree_boundaries[0], cross_fields_boundaries[0]
        self.calc_cross_boundaries(cross_fields_bins)
        df_X = self.get_cross_features(df_X)
        for icol in df_X.columns:
            if 'ext-' not in icol:
                continue
            else:
                value_ = [-np.inf,np.inf]
                value_.extend(df_X[icol].unique()[1:-1])
                value_.sort()
                tree_bins.update({icol:value_})
        base_woe, x = woe.get_woe_card(df_X, df_Y, tree_bins)
        clf_lr = LogisticRegression(**self.params_lr)
        clf_lr.fit(x, df_y_binary, sample_weight=sample_weight)
        clf_lr.col_name = x.columns

        predict_ = clf_lr.predict_proba(x)[:, 1:].sum(axis=1)
        train_auc = self.score_model.get_auc(pd.DataFrame(predict_), df_y_binary.copy(), pre_target=1)[0]
        log_step("Init train_auc={}".format(round(train_auc, 4)))

        base_card = self.get_score_card(clf_lr, base_woe)
        self.init_cards.append(base_card)
        return df_X

    def calc_cross_boundaries(self, boundaries):
        '''

        :param boundaries: 默认按重要度由大到小排序的字段分箱
        :return:
        '''
        if len(boundaries) < 1:
            return
        init_ext = self.feature_ext_info.get('df_ext',pd.DataFrame())
        init_boundaries = self.feature_ext_info.get('boundaries_ext',{})
        boundaries.update(init_boundaries)
        df_ext = pd.DataFrame(columns=['id_'])
        for col_, bins_ in boundaries.items():
            df_tmp = pd.cut([], bins=bins_).value_counts().sort_index().reset_index().rename(
                columns={'index': col_, 0: 'id_'})
            df_ext = df_ext.merge(df_tmp, how='outer', on='id_')
        df_ext['id_'] = range(len(df_ext))

        for i, j in itertools.product(boundaries.keys(), boundaries.keys()):
            if i==j:
                continue
            if i not in df_ext.columns:
                continue
            if j not in df_ext.columns:
                continue
            col_='ext-{}-{}'.format(i,j)
            if col_ in init_ext.columns:
                df_tmp = init_ext.copy()
                df_tmp['tmp_'] = df_tmp[[i, j]].astype(str).apply(lambda x: '-'.join(x), axis=1)
                df_tmp = df_tmp[['tmp_',col_]].drop_duplicates()
            else:
                df_tmp = pd.DataFrame(df_ext[[i,j]].astype(str).apply(lambda x: '-'.join(x), axis=1).drop_duplicates())
                df_tmp.columns=['tmp_']
                df_tmp[col_] = range(len(df_tmp))
            df_ext['tmp_'] = df_ext[[i,j]].astype(str).apply(lambda x: '-'.join(x), axis=1)
            df_ext = df_ext.join(df_tmp.set_index('tmp_'),on='tmp_',how='inner')
        if 'tmp_' in df_ext.columns:
            del df_ext['tmp_']
        self.feature_ext_info = {'df_ext':df_ext,'boundaries_ext':boundaries}

    def get_cross_features(self, df_data):
        df_X_ext = self.feature_ext_info.get('df_ext')
        boundaries_ext = self.feature_ext_info.get('boundaries_ext')
        if not(boundaries_ext):
            return df_data
        if len(boundaries_ext)<2:
            return df_data
        df_X = df_data.copy()
        df_tmp = pd.DataFrame()
        for k,bv in boundaries_ext.items():
            df_tmp[k]=pd.cut(df_X[k],bins=bv).astype(str)
        for icol in df_X_ext.columns:
            if 'ext-' not in icol:
                continue
            col_ = icol[4:].split('-')
            df_tmp['tmp_'] = df_tmp[col_].apply(lambda x: '-'.join(x), axis=1)
            df_X_ext['tmp_'] = df_X_ext[col_].astype(str).apply(lambda x: '-'.join(x), axis=1)
            df_ext_tmp = df_X_ext[['tmp_',icol]].drop_duplicates()
            df_tval = pd.merge(df_tmp,df_ext_tmp.set_index('tmp_'),how='inner',left_on='tmp_',right_index=True)[icol]
            assert len(df_tval)==len(df_X), (len(df_tval),len(df_X))
            df_X[icol] = df_tval
        return df_X

    def get_score_card(self,clf_lr,df_woe):
        coef_dict = dict(zip(clf_lr.col_name, clf_lr.coef_[0]))
        init_score = self.score_model.score_offset
        base_score = self.score_model.score_factor * clf_lr.intercept_[0]
        woe_col = [i for i in df_woe.columns if 'woe' in i]
        get_coef_summary = lambda x: np.sum([coef_dict.get(x['field_'] + '_' + i) * x[i] for i in woe_col])
        df_woe['score_'] = self.score_model.score_factor * df_woe.apply(get_coef_summary, axis=1)
        df_field_card = df_woe[['field_', 'bins_', 'boundary_', 'size_', 'score_']]
        df_base_card = pd.DataFrame(
            [['init_base_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1, init_score],
             ['init_model_score', pd.Interval(-np.inf, np.inf, closed='right'),np.inf,  -1, base_score]
             ], columns=['field_', 'bins_', 'boundary_', 'size_', 'score_'])
        df_card = df_field_card.append(df_base_card, ignore_index=True)
        return df_card

    def get_ml_boundaries(self, df_X, df_y, sample_weight):
        if self.features_model =='rf':
            clf_rf = RandomForestClassifier(**self.params_rf)
            clf_rf.fit(df_X, df_y, sample_weight=sample_weight)
            rf_report = classification_report(df_y, clf_rf.predict(df_X))
            log_step("\nRF-report: \n {}".format(rf_report))
            rf_boundaries, _ = rf_bolcks.get_randomforest_blocks(clf_rf, col_name=df_X.columns)
            return rf_boundaries
        if self.features_model =='xgb':
            clf_xgb = XGBClassifier(**self.params_xgb)
            clf_xgb.fit(df_X, df_y, sample_weight=sample_weight)
            xgb_report = classification_report(df_y, clf_xgb.predict(df_X))
            log_step("\nXGB-report: \n {}".format(xgb_report))
            xgb_boundaries = xgb_bolcks.get_xgboost_blocks(clf_xgb, col_name=df_X.columns)
            return xgb_boundaries

    def get_init_ecard(self, df_X, ext_boundaries_list):
        boundaries_list = []
        if len(ext_boundaries_list)>0:
            boundaries_list.extend(ext_boundaries_list)
        for icard in self.init_cards:
            base_date=icard.groupby('field_').boundary_.agg(lambda x:sorted(set([-np.inf,np.inf]+list(x))) if len(x)>2 else None)
            base_boundaries = base_date[base_date.notna()].to_dict()
            if len(base_boundaries) > 0:
                boundaries_list.append(base_boundaries)
        boundaries = self.get_boundaries_merge(boundaries_list)
        self.blocks = boundaries

        len_=len(boundaries)
        bins_list=[]
        df_X_bins = pd.DataFrame()
        for i, (col, bins) in enumerate(boundaries.items()):
            progress_bar(i,len_-1)
            if col not in df_X.columns:
                log_step(" ------ {} column not exists, ignore!")
                continue
            idata = pd.cut(df_X.loc[:, col].fillna(0), bins=bins)
            df_X_bins[col] = idata
            data_bin = idata.value_counts()
            df_bin = pd.DataFrame(data_bin).sort_index().reset_index()
            df_bin.columns = ['bins_','size_']
            df_bin.insert(loc=0, column='field_', value=col)
            df_bin.insert(loc=2, column='boundary_', value=bins[1:])
            bins_list.append(df_bin)
        df_init_ecard = pd.concat(bins_list).reset_index(drop=True)
        df_base_card = pd.DataFrame(
            [['init_base_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf,-1],
             ['init_model_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1]
             ], columns=['field_', 'bins_', 'boundary_', 'size_'])
        df_card = df_init_ecard.append(df_base_card, ignore_index=True)
        df_card['score_']=0

        cur_card = df_card
        for icard_ in self.init_cards:
            if self.cur_card_index==0:
                cur_card = self.get_cards_merge(cur_card, icard_, cur_weight=0, add_weight=1.0)
            else:
                cur_card = self.get_cards_merge(cur_card, icard_,
                                                cur_weight=self.cur_card_index / (self.cur_card_index + 1),
                                                add_weight=1.0 / (self.cur_card_index + 1))
            self.cur_card_index += 1

        return cur_card, df_X_bins

    def get_cards_merge(self, cur_card, add_card, cur_weight=0.5, add_weight=0.5):
        df_cur = cur_card.set_index(['field_', 'boundary_', 'size_'])
        df_add = cur_card.join(add_card.set_index('field_')[['bins_', 'score_']], on='field_', how='left',
                                rsuffix='_')
        idx = df_add.apply(
            lambda x: True if str(x['bins__']) != 'nan' and (x['bins_'].overlaps(x['bins__'])) else False, axis=1)
        df_add.loc[~idx, 'score__'] = 0
        df_add = df_add.groupby(['field_', 'boundary_', 'size_'])['score__'].sum()
        df_cur['score_'] = df_cur['score_']*cur_weight + df_add*add_weight
        return df_cur.reset_index()

    def get_boundaries_merge(self, boundaries_list):
        gl_boundaries={}
        for boundaries in boundaries_list:
            for k,v in boundaries.items():
                if k in gl_boundaries:
                    lv = gl_boundaries.get(k)
                    cv = list(set(v+lv))
                    cv.sort()
                    gl_boundaries.update({k:cv})
                else:
                    gl_boundaries.update({k: v})
        return gl_boundaries

    def calc_score(self, df_card, df_data: pd.DataFrame, is_woe_feature=False):
        df_score = pd.DataFrame()
        if is_woe_feature:
            pass
        else:
            bins_dict = df_card.groupby('field_').boundary_.agg(
                lambda x: sorted(set([-np.inf, np.inf] + list(x)))).to_dict()
        for field_ in df_card.field_.unique():
            if field_ in ['init_base_score','init_model_score']:
                continue
            _card = df_card[df_card.field_ == field_]
            _card["bins_"] = _card.bins_.astype(str)
            if is_woe_feature:
                _data = df_data[field_].astype(str)
            else:
                _data = pd.cut(df_data[field_].fillna(0), bins=bins_dict.get(field_, [-np.inf, np.inf])).astype(str)
            out = pd.merge(_data,_card.set_index("bins_"),how='left',left_on=_data.name,right_index=True)
            df_score[field_] = out['score_']
        df_score['init_base_score'] = df_card[df_card.field_ == 'init_base_score']['score_'].values[0]
        df_score['init_model_score'] = df_card[df_card.field_ == 'init_model_score']['score_'].values[0]
        score = df_score.sum(axis=1).apply(lambda x: round(x,2))
        return score

    def get_importance_(self):
        df_card_info = self.score_ecard.groupby("field_").agg(
            {'size_': [("weights", lambda x: list(x))],
             'score_': [("values", lambda x: list(x))]

             }).droplevel(0, 1)
        importance_ = df_card_info.apply(lambda x: get_weighted_std(x["values"], x["weights"]), axis=1)
        importance_ = (importance_/importance_.sum()).apply(lambda x:round(x,6))
        return importance_.sort_values(ascending=False)

    def predict(self, df_data):
        if self.cross_hierarchy>1:
            df_data = self.get_cross_features(df_data)
        df_score = self.calc_score(self.score_ecard,df_data)
        return df_score

    def predict_proba(self, df_data):
        df_score = self.predict(df_data)
        df_proba =df_score.apply(lambda x:self.score_model.score_to_probability(x))
        return df_proba

    def predict_hundred(self, df_data):
        '''
        百分制
        :param df_data:
        :return:
        '''
        df_score = self.predict(df_data)
        df_hundred =df_score.apply(lambda x:round((x-self.score_lower)/(self.score_upper-self.score_lower)*100,2))
        return df_hundred

    # 单车详细评分
    def get_single_score(self, data: dict, level_threshold=None):
        data = self.get_ext_features(data)
        score = 0
        score_detail = {}
        df_card = self.score_ecard
        for i, row in df_card.iterrows():
            field_ = row['field_']
            bins_ = row['bins_']
            score_ = row['score_']
            # class_ = row['class_']
            if field_ in ['init_base_score', 'init_model_score']:
                v = 0
            else:
                v = data.get(field_)
            try:
                v = float(v)
            except:
                score = 0
                score_detail[field_] = '异常'
                break
            if not (v > -np.inf):
                v = 0
            if v in bins_:
                score = score + score_
                score_detail[field_] = score_
            else:
                pass

        # level
        score = round(score)
        level = -1
        total_meters = data.get('run_meters', 0)
        total_seconds = data.get('run_seconds', 0)
        if (level_threshold is not None) and (total_meters >= 2000000) and (total_seconds >= 10 * 3600):
            level = np.argwhere(np.sort(level_threshold + [score]) == score).min()

        out = {'score': score, 'level': level, 'score_detail': score_detail}
        return out

    def get_ext_features(self, data):
        if hasattr(self, 'feature_ext_info'):
            df_ext = self.feature_ext_info.get("df_ext")
            if type(df_ext)!=pd.DataFrame:
                return data
            if len(df_ext)<1:
                return data
            ext_columns = [i for i in df_ext.columns if 'ext-' in i]
            for i in ext_columns:
                colinfo = str(i).split('-')
                if 'ext' == colinfo[0]:
                    idx_equ = True
                    for icol in colinfo[1:]:
                        if icol not in df_ext.columns:
                            break
                        idx_equ = idx_equ & df_ext[icol].apply(lambda x: data.get(icol) in x).astype(bool)
                    value_ = df_ext.loc[idx_equ, i].values
                    if len(value_) > 0:
                        data.update({i: value_[0]})
                    else:
                        data.update({i: 0})
        return data

    def init_check(self):
        if isinstance(self.init_cards,pd.DataFrame):
            col_ = [i for i in self.init_cards.columns if 'ext-'==i[:4]]
            if len(col_)>0:
                df_ext = self.feature_ext_info.get('df_ext')
                for i in col_:
                    if i not in df_ext.columns:
                        log_step("no detail info for cross field:{}".format(i))
                        break

    def model_explanation(self, title=None, save_path=None,**kwargs):
        return summary_explanation(self.score_ecard, title=title, save_path=save_path,**kwargs)

    def factor_report(self, title='ecard', max_display=999, save_path=None):
        return model_factor_report(self.score_ecard, title, max_display, save_path)

    def dump(self, model_name, score_bins):
        self.score_bins = score_bins
        ecard_standardized = ECardStandardized()
        ecard_standardized.dump(self, model_name, model='all')

if __name__ == '__main__':
    df_valid = pd.read_csv("data/train_test_data.csv")
    df_train_data = df_valid[df_valid['train_test_tag'] == '训练集'].fillna(0).head(1000)
    df_test_data = df_valid[df_valid['train_test_tag'] == '测试集'].fillna(0).head(1000)
    feature_columns = df_train_data.columns[4:33].tolist()
    feature_columns.extend(df_train_data.columns[36:46])
    df_X = df_train_data[feature_columns]
    df_Y = df_train_data[['label', 'label_ordinary',
                          'label_serious', 'label_major', 'label_devastating', 'label_8w','fee_got','report_fee']]
    df_Y['label']=df_Y.apply(lambda x:x['label'] if x["report_fee"]<5000 else 2,axis=1)
    ecard = ECardModel(
        kwargs_rf={'n_estimators':20},
        kwargs_xgb={'n_estimators':2},
        cross_hierarchy=3,
        is_best_bagging=True,
        features_model='xgb',
        optimize_type='local',
    )


    # 初始加载河北模型进行微调 -------------------
    import dill
    def load_local_model(model_path):
        model = dill.load(open(model_path, "rb"))
        clf = model.get('model')
        score_card = model.get('score_card')
        score_bins = model.get('score_bins')
        get_card_level = model.get('get_card_level')
        func_calc_level = lambda x: get_card_level(x["card_score"], x["run_meters"], x["run_seconds"], score_bins)

        return clf, func_calc_level


    model_file = 'data/score_card_model_local_major_hebeibeijing_v3.1.0.pkl'
    clf_hebei, func_calc_level = load_local_model(model_path=model_file)
    ecard.init_cards.append(clf_hebei.score_ecard)
    ecard.feature_ext_info = clf_hebei.feature_ext_info
    ### ----------------------------------------------

    ecard.fit(df_X, df_Y,df_X, df_Y,sample_weight=df_Y['label']+1, validation_step=1)
    print(ecard.get_importance_())
    print(ecard.predict(df_X))
    print(ecard.predict_hundred(df_X))
    data = df_test_data.loc[0].to_dict()
    print(ecard.get_single_score(data=data, level_threshold=[-np.inf,400,536,550,600,np.inf]))
    print("""
    3.calc score 优化为predict，提高速度
    4.woe计算优化
    5.并行优化
    """)


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/5/6 18:29
Desc:
'''
import pandas as pd
from score_ecard.features.xgboost_woe import XGBoostWoe

if __name__ == '__main__':
    df_valid = pd.read_csv("../score_ecard/data/train_test_data.csv")
    df_train_data = df_valid[df_valid['train_test_tag'] == '训练集'].fillna(0).head(100000)
    df_test_data = df_valid[df_valid['train_test_tag'] == '测试集'].fillna(0).head(100000)
    feature_columns = df_train_data.columns[4:33].tolist()
    feature_columns.extend(df_train_data.columns[36:46])
    df_X = df_train_data[feature_columns]
    df_Y = df_train_data[['label', 'label_ordinary',
                          'label_serious', 'label_major', 'label_devastating', 'label_8w', 'fee_got', 'report_fee']]
    df_Y = df_Y.apply(lambda x: x['label'] if x["report_fee"] < 5000 else 1, axis=1)*2

    xgb_params = {
            'n_estimators': 2,
            'max_depth':5,
            'eta': 0.3,
            'min_child_weight': 1,
            'gamma': 0.3,
            'subsample': 0.8,
            'colsample_bytree': 0.8,
            'booster': 'gbtree',
            'objective': 'binary:logistic',
            'scale_pos_weight': 1,
            'lambda': 1,
            'seed':666,
            'silent': 0,
            'eval_metric': 'auc'
        }
    xgb_woe = XGBoostWoe(xgb_params)
    df_woe_list=xgb_woe.fit(df_X,df_Y)
    bb=xgb_woe.transform(df_X)

    aa=pd.concat(df_woe_list,axis=1)
    print(aa.columns)


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/6 00:11
Desc:
'''


import numpy as np
import pandas as pd

## 评分体系
class ScoreCardModel:
    def __init__(self):
        self.load_score_alpha()

    def load_score_alpha(self, odds=1 / 2, bScore=600, addScore=-30):
        '''
            bScore = offset - factor*ln(odds)
            bScore + addScore = offset - factor*ln(2*odds)
        :param odds: 坏好比
        :param bScore: 基础score分
        :param addScore: 评分变动值
        :return: bScore = offset - factor*ln(odds)
        '''

        self.score_factor = addScore / np.log(2)
        self.score_offset = bScore - addScore * np.log(odds) / np.log(2)

    def probability_to_score(self, bad_prob, eps=1e-5):
        '''
            逾期概率转换为评分
        :param bad_prob: 逾期概率
        :return: bScore = offset - factor*ln(odds)
        '''

        odds = max(bad_prob, eps) / max(1 - bad_prob, eps)
        score = self.score_offset + self.score_factor * np.log(odds)
        return int(round(score))

    def score_to_probability(self, score, eps=1e-5):
        '''
            评分转换为逾期概率
        :param score: 信用评分
        :return: bScore = offset - factor*ln(odds) -->bad_prob
        '''
        score_factor = eps if self.score_factor==0 else self.score_factor
        odds = np.e**((score - self.score_offset)/(score_factor))
        odds = -1+eps if odds == -1 else odds
        bad_prob = 1.00*odds/(1+odds)
        return bad_prob

    @classmethod
    def get_auc(cls, df_pre, df_label, df_weight=None, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签{0,1}
        :param df_weight: 一维数组或series，代表样本权重
        :param pre_target: 'auc': auc值，'crossdens': TPR&FPR
        :return:
        '''
        if isinstance(df_pre, pd.DataFrame):
            df_pre = df_pre.iloc[:,0]
        if isinstance(df_label, pd.DataFrame):
            df_label = df_label.iloc[:,0]
        if isinstance(df_weight, pd.DataFrame):
            df_weight = df_weight.iloc[:,0]

        if type(df_weight)==type(None):
            df_weight = df_pre*np.array(0)+1
        assert len(df_pre) == len(df_weight) == len(df_label)
        df_pre.reset_index(drop=True, inplace=True)
        df_weight.reset_index(drop=True, inplace=True)
        df_label.reset_index(drop=True, inplace=True)

        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        if df_label.unique().size < 2:
            return None, None
        # 按照预测概率从大到小排序作为阈值
        crossfreq = pd.crosstab(df_pre, df_label, df_weight, aggfunc=sum).fillna(0).sort_index(ascending=False)
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        auc = np.trapz(y=tpr, x=fpr)
        return auc, crossdens

    @classmethod
    def get_insurance_auc(cls, df_pre, df_got_fee, df_report_fee, df_weight=None):
        '''
        功能: 计算auc值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_got_fee: 一维数组或series，代表已赚保费
        :param df_report_fee: 一维数组或series，代表赔付额
        :param df_weight:
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        if isinstance(df_pre, pd.DataFrame):
            df_pre = df_pre.iloc[:,0]
        if isinstance(df_got_fee, pd.DataFrame):
            df_got_fee = df_got_fee.iloc[:,0]
        if isinstance(df_report_fee, pd.DataFrame):
            df_report_fee = df_report_fee.iloc[:,0]
        if isinstance(df_weight, pd.DataFrame):
            df_weight = df_weight.iloc[:,0]

        if type(df_weight)==type(None):
            df_weight = df_pre*np.array(0)+1
        assert len(df_pre) == len(df_weight) == len(df_got_fee) == len(df_report_fee)
        df_pre.reset_index(drop=True, inplace=True)
        df_weight.reset_index(drop=True, inplace=True)
        df_got_fee.reset_index(drop=True, inplace=True)
        df_report_fee.reset_index(drop=True, inplace=True)

        df_data = pd.concat([df_got_fee, df_report_fee],axis=1, ignore_index=True)
        df_data = df_data.mul(df_weight,axis=0).fillna(0)
        df_data.columns = ['got_fee', 'report_fee']
        order_index = df_pre.reset_index(drop=True).sort_values(ascending=False).index
        crossdens = df_data.loc[order_index,:].cumsum(axis=0) / df_data.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        auc = np.trapz(y=tpr, x=fpr)
        return auc, crossdens

    @classmethod
    def get_ks(cls, df_pre, df_label, df_weight=None, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签
        :param df_weight: 一维数组或series，代表样本权重
        :param pre_target:
        :return:
            'ks': KS值
            'crossdens': 好坏客户累积概率分布以及其差值gap
        '''
        if isinstance(df_pre, pd.DataFrame):
            df_pre = df_pre.iloc[:, 0]
        if isinstance(df_label, pd.DataFrame):
            df_label = df_label.iloc[:, 0]
        if isinstance(df_weight, pd.DataFrame):
            df_weight = df_weight.iloc[:, 0]

        if type(df_weight) == type(None):
            df_weight = df_pre * np.array(0) + 1
        assert len(df_pre) == len(df_weight) == len(df_label)
        df_pre.reset_index(drop=True, inplace=True)
        df_weight.reset_index(drop=True, inplace=True)
        df_label.reset_index(drop=True, inplace=True)

        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        if df_label.unique().size < 2:
            return None, None
        # 按照预测概率从大到小排序
        crossfreq = pd.crosstab(df_pre, df_label, df_weight, aggfunc=sum).fillna(0).sort_index(ascending=False)
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        crossdens['gap'] = abs(crossdens['fpr'] - crossdens['tpr'])
        ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
        return ks, crossdens

    @classmethod
    def get_insurance_ks(cls, df_pre, df_got_fee, df_report_fee, df_weight=None):
        '''
        功能: 计算auc值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_got_fee: 一维数组或series，代表已赚保费
        :param df_report_fee: 一维数组或series，代表赔付额
        :param df_weight:
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        if isinstance(df_pre, pd.DataFrame):
            df_pre = df_pre.iloc[:,0]
        if isinstance(df_got_fee, pd.DataFrame):
            df_got_fee = df_got_fee.iloc[:,0]
        if isinstance(df_report_fee, pd.DataFrame):
            df_report_fee = df_report_fee.iloc[:,0]
        if isinstance(df_weight, pd.DataFrame):
            df_weight = df_weight.iloc[:,0]

        if type(df_weight)==type(None):
            df_weight = df_pre * np.array(0) + 1
        assert len(df_pre) == len(df_weight) == len(df_got_fee) == len(df_report_fee)
        df_pre.reset_index(drop=True, inplace=True)
        df_weight.reset_index(drop=True, inplace=True)
        df_got_fee.reset_index(drop=True, inplace=True)
        df_report_fee.reset_index(drop=True, inplace=True)

        df_data = pd.concat([df_got_fee, df_report_fee],axis=1, ignore_index=True)
        df_data = df_data.mul(df_weight,axis=0).fillna(0)
        df_data.columns = ['got_fee', 'report_fee']
        order_index = df_pre.reset_index(drop=True).sort_values(ascending=False).index
        crossdens = df_data.loc[order_index,:].cumsum(axis=0) / df_data.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        crossdens['gap'] = abs(crossdens['fpr'] - crossdens['tpr'])
        ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
        return ks, crossdens

    @classmethod
    def get_auc_report(cls, df_pre, df_label, df_report_fee, df_got_fee=None, df_subset_identity=None, df_weight=None,
                       pre_target=1, subset_weight='balance'):

        if type(df_got_fee) == type(None):
            df_got_fee = df_pre.apply(lambda _: 15000)
        if type(df_subset_identity) == type(None):
            df_subset_identity = df_pre.apply(lambda _: '全量')
        if type(df_weight) == type(None):
            if 'balance' == subset_weight:
                dict_ = (df_subset_identity.shape[0] / df_subset_identity.value_counts()).to_dict()
                dict_.update({"其他": 1})
                df_weight = df_subset_identity.apply(lambda x: dict_.get(x, 1))
            else:
                df_weight = df_pre * np.array(0) + 1
        assert len(df_pre) == len(df_got_fee) == len(df_report_fee) == len(df_subset_identity) == len(df_weight), print(
            len(df_pre), len(df_got_fee), len(df_report_fee), len(df_subset_identity), len(df_weight))
        df_pre.reset_index(drop=True, inplace=True)
        df_got_fee.reset_index(drop=True, inplace=True)
        df_report_fee.reset_index(drop=True, inplace=True)
        df_weight.reset_index(drop=True, inplace=True)
        df_subset_identity.reset_index(drop=True, inplace=True)

        auc_report_all = cls.get_auc(df_pre, df_label, df_weight, pre_target)[0]
        auc_fee_all = cls.get_insurance_auc(df_pre, df_got_fee, df_report_fee, df_weight)[0]
        index_ = ['全量']
        data_ = [[len(df_pre), auc_report_all, auc_fee_all]]

        if df_subset_identity.unique().size > 1:
            for ii in df_subset_identity.unique():
                idx = df_subset_identity[df_subset_identity == ii].index
                auc_report = cls.get_auc(df_pre.loc[idx], df_label.loc[idx], df_weight.loc[idx], pre_target)[0]
                auc_fee = \
                cls.get_insurance_auc(df_pre.loc[idx], df_got_fee.loc[idx], df_report_fee.loc[idx], df_weight.loc[idx])[0]
                data_.append([len(idx), auc_report, auc_fee])
                index_.append(ii)
        df_report = pd.DataFrame(data_, columns=['样本量', '出险率_auc', '赔付率_auc'], index=index_).sort_values("样本量",
                                                                                                         ascending=False)
        return df_report

    @classmethod
    def get_auc_report2(cls, df, by_field='subset_identity', weight_field=None, pre_target=1, subset_weight='balance',
                        fix_fee_got=True):
        assert 'predict' in df.columns
        assert 'label' in df.columns
        assert 'fee_got' in df.columns
        assert 'report_fee' in df.columns
        df_pre = df.predict
        df_label = df.label
        df_got_fee = df.fee_got
        df_report_fee = df.report_fee
        if fix_fee_got:
            df_got_fee = df_got_fee * 0 + 1000
        if by_field in df.columns:
            df_subset_identity = df[by_field]
        else:
            df_subset_identity = None
        if weight_field in df.columns:
            df_weight = df[weight_field]
        else:
            df_weight = None
        df_report = cls.get_auc_report(df_pre, df_label, df_got_fee, df_report_fee, df_subset_identity, df_weight,
                                   pre_target, subset_weight)

        return df_report

# 评分校准
def get_calibration_score(raw_score):
    pass

if __name__ == '__main__':
    model_score = ScoreCardModel()
    score = model_score.probability_to_score(0.5)
    print(score)


import setuptools
from version import get_git_version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
	name='score_ecard',
	version=get_git_version(),
	description='random forest + lr card',
	url='https://github.com/wqwangchn/score_ecard',
	author='wqwangchn',
	author_email='wqwangchn@163.com',
	long_description=long_description,
	long_description_content_type="text/markdown",
	packages=setuptools.find_packages(),
	include_package_data=True,
	install_requires=['pandas','numpy','scipy','interval'],
	classifiers=[
	"Programming Language :: Python :: 3",
	"License :: OSI Approved :: MIT License",
	"Operating System :: OS Independent",
	],
	zip_safe=False
)



# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/4/28 15:41
Desc:
'''

import numpy as np
import pandas as pd
import math
import random
import warnings
warnings.filterwarnings('ignore')

from sklearn.linear_model import LogisticRegression
from score_ecard import score_card
from score_ecard.util import get_weighted_std, log_step,summary_explanation,model_factor_report
from score_ecard.features.xgboost_woe import XGBoostWoe
from score_ecard.util.ecard_standardized import ECardStandardized

class ECardModel(object):
    def __init__(self, kwargs_lr=None, kwargs_xgb=None, sample_frac=1.0, cross_hierarchy=0, n_estimators=None, **kwargs):
        '''

        :param kwargs_lr:
        :param kwargs_xgb:
        :param sample_frac: 样本再抽样比例，样本抽样比例 = 0.6*sample_frac
        :param cross_hierarchy: 特征交叉层级，可参考参数设置为[2，3]，默认不进行特征扩展即cross_hierarchy=0
        '''
        self.params_lr = {
            "class_weight": 'balanced',
            "penalty": 'l1',
            "C": 0.2,
            "solver": 'liblinear'
        }
        self.params_xgb = {
            'n_estimators': 10,
            'max_depth':5,
            'learning_rate': 0.3,
            'min_child_weight': 1,
            'gamma': 0.3,
            'subsample': 0.8,
            'colsample_bytree': 0.8,
            'booster': 'gbtree',
            'objective': 'binary:logistic',
            'scale_pos_weight': 1,
            'lambda': 1,
            'seed':666,
            'silent': 0,
            'eval_metric': 'auc'
        }
        self.params_base_tree = {
            'max_depth': 3,
            'max_features': None,
            'min_samples_leaf': 0.01,
            'class_weight': "balanced",
            'random_state': 666
        }
        if kwargs_lr:
            self.params_lr.update(**kwargs_lr)
        if kwargs_xgb:
            self.params_xgb.update(**kwargs_xgb)
        self.sample_frac = sample_frac
        self.cross_hierarchy = cross_hierarchy
        self.score_ecard = None
        self.score_model = score_card.ScoreCardModel()
        random.seed(666)
        if n_estimators is None:
            self.n_estimators = self.params_xgb.get("n_estimators",100)
        else:
            self.n_estimators = n_estimators

        self.xgb_woe_model = XGBoostWoe(self.params_xgb)
        self.init_cards=[]
        self.feature_ext_info={}

    def fit(self, df_X, df_Y, sample_weight=None, df_valX:pd.DataFrame=None, df_valY:pd.DataFrame=None):
        assert df_X.shape[0] == df_Y.shape[0]
        df_y, df_y_binary= self.check_label(df_Y)
        if type(sample_weight)==type(None):
            sample_weight=df_y*0+1
        log_step("start model fitting ...")
        df_woe_list = self.xgb_woe_model.fit(df_X, df_y)
        df_cards_list = self.calc_score_card_v2(df_woe_list, df_y_binary, sample_weight)
        self.score_ecard = self.merge_cards(df_cards_list,df_X)

        self.score_lower = self.score_ecard.groupby("field_").score_.min().sum()
        self.score_upper = self.score_ecard.groupby("field_").score_.max().sum()

    def check_label(self, df_label):
        df_Y=df_label.copy()
        if df_Y.shape[-1] == 1:
            df_Y.columns = ['label']
        if 'label' not in df_Y.columns:
            df_Y['label'] = df_Y.iloc[:,0]
        df_y = df_Y.label.astype(int)
        if df_y.unique().size>2:
            df_y_binary = df_y.apply(lambda x: 1 if x>0 else 0)
        else:
            df_y_binary = df_y
        ## test---
        if 'label_binary' in df_Y.columns:
            df_y_binary = df_Y["label_binary"]
        ## ---end
        del df_Y
        return df_y, df_y_binary

    def calc_score_card(self, df_woe_list, df_y_binary, sample_weight):
        '''
        pass
        :param df_woe_list:
        :param df_y_binary:
        :param sample_weight:
        :return:
        '''
        woe_cards_list = self.xgb_woe_model.woe_cards.copy()
        col_list = [i.columns for i in df_woe_list]
        df_woe_data = pd.concat(df_woe_list,axis=1)

        clf_lr = LogisticRegression(**self.params_lr)
        clf_lr.fit(df_woe_data, df_y_binary, sample_weight=sample_weight)
        del df_woe_data

        # score_card of xgb_trees
        lr_coef = clf_lr.coef_[0]
        i = 0
        for idx_,(icol_,df_iwoe) in enumerate(zip(col_list,woe_cards_list)):
            icol_coef_ = lr_coef[i:i+len(icol_)]
            coef_dict = dict(zip(icol_, icol_coef_))

            woe_col = [i for i in df_iwoe.columns if 'woe' in i]
            get_coef_summary = lambda x: np.sum([coef_dict.get(x['field_'] + '_' + i) * x[i] for i in woe_col])
            df_iwoe['score_'] = self.score_model.score_factor * df_iwoe.apply(get_coef_summary, axis=1)
            woe_cards_list[idx_] = df_iwoe[['field_', 'bins_', 'boundary_', 'size_', 'score_']]
            i += len(icol_)

        #  init score_card
        init_score = self.score_model.score_offset
        base_score = self.score_model.score_factor * clf_lr.intercept_[0]
        df_base_card = pd.DataFrame(
            [['init_base_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1, init_score],
             ['init_model_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1, base_score]
             ], columns=['field_', 'bins_', 'boundary_', 'size_', 'score_'])
        woe_cards_list.append(df_base_card)

        return woe_cards_list

    def calc_score_card_v2(self, df_woe_list, df_y_binary, sample_weight,batch_zise=100):
        '''

        :param df_woe_list:
        :param df_y_binary:
        :param sample_weight:
        :return:
        '''
        woe_cards_list = self.xgb_woe_model.woe_cards.copy()
        clf_lr = LogisticRegression(**self.params_lr)
        batch_num = math.ceil(len(df_woe_list) / batch_zise)
        col_list = [i.columns for i in df_woe_list]
        if batch_num<=1:
            df_woe_data = pd.concat(df_woe_list,axis=1)
            del df_woe_list
            clf_lr.fit(df_woe_data, df_y_binary, sample_weight=sample_weight)
            del df_woe_data
            lr_coef = clf_lr.coef_[0]
            lr_intercept_ = clf_lr.intercept_[0]
        else:
            lr_coef_list = []
            lr_intercept_list = []
            lf_score_list = []
            for i in range(batch_num):
                df_idata = pd.concat(df_woe_list[:batch_num],axis=1)
                del df_woe_list[:batch_num]
                clf_lr.fit(df_idata, df_y_binary, sample_weight=sample_weight)
                lr_coef_list.append(clf_lr.coef_[0])
                lr_intercept_list.append(clf_lr.intercept_[0])
                lr_iscore = clf_lr.decision_function(df_idata)
                lf_score_list.append(lr_iscore)
                del df_idata
            lf_score_ = np.array(lf_score_list).T

            clf_lr2 = LogisticRegression(penalty='l2', C=1.0, class_weight=True, random_state=666)
            clf_lr2.fit(lf_score_, df_y_binary, sample_weight=sample_weight)
            for i,w in enumerate(clf_lr2.coef_[0]):
                lr_coef_list[i] = lr_coef_list[i]*w
                lr_intercept_list[i] = lr_intercept_list[i]*w
            lr_coef = np.concatenate(lr_coef_list)
            lr_intercept_ = sum(lr_intercept_list)+clf_lr2.intercept_[0]

        # score_card of xgb_trees
        i = 0
        for idx_,(icol_,df_iwoe) in enumerate(zip(col_list,woe_cards_list)):
            icol_coef_ = lr_coef[i:i+len(icol_)]
            coef_dict = dict(zip(icol_, icol_coef_))

            woe_col = [i for i in df_iwoe.columns if 'woe' in i]
            get_coef_summary = lambda x: np.sum([coef_dict.get(x['field_'] + '_' + i) * x[i] for i in woe_col])
            df_iwoe['score_'] = self.score_model.score_factor * df_iwoe.apply(get_coef_summary, axis=1)
            woe_cards_list[idx_] = df_iwoe[['field_', 'bins_', 'boundary_', 'size_', 'score_']]
            i += len(icol_)

        #  init score_card
        init_score = self.score_model.score_offset
        base_score = self.score_model.score_factor * lr_intercept_
        df_base_card = pd.DataFrame(
            [['init_base_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1, init_score],
             ['init_model_score', pd.Interval(-np.inf, np.inf, closed='right'), np.inf, -1, base_score]
             ], columns=['field_', 'bins_', 'boundary_', 'size_', 'score_'])
        woe_cards_list.append(df_base_card)

        return woe_cards_list

    def merge_cards(self, woe_cards_list,df_trian=None):
        cards=pd.concat(woe_cards_list,axis=0).reset_index(drop=True)
        cards_boundaries = cards.groupby('field_').boundary_.agg(lambda x: sorted(set([-np.inf, np.inf] + list(x))))

        if type(df_trian)==type(None):
            df_tmp = pd.DataFrame(cards_boundaries).applymap(lambda x: 0).T
        else:
            df_tmp = df_trian.copy()
            df_tmp['init_base_score']=0
            df_tmp['init_model_score'] = 0

        bins_list = []
        for i, (col, bins) in enumerate(cards_boundaries.items()):
            data_bin = pd.cut(df_tmp.loc[:, col], bins=bins).value_counts()
            df_bin = pd.DataFrame(data_bin).sort_index().reset_index()
            df_bin.columns = ['bins_', 'size_']
            df_bin.insert(loc=0, column='field_', value=col)
            df_bin.insert(loc=2, column='boundary_', value=bins[1:])
            bins_list.append(df_bin)
        df_init_ecard = pd.concat(bins_list).reset_index(drop=True)
        del df_tmp

        df_cards_out = pd.merge(df_init_ecard,cards.set_index('field_')[['bins_', 'size_','score_']],how='left',left_on='field_',right_index=True, suffixes=("", "y"))
        idx = df_cards_out.apply(
            lambda x: True if str(x['bins_y']) != 'nan' and (x['bins_'].overlaps(x['bins_y'])) else False, axis=1)
        df_cards_out.loc[~idx, 'score_'] = 0
        df_cards_out.loc[~idx, 'size_y'] = 0
        if type(df_trian)==type(None):
            df_cards_out['size_']= df_cards_out['size_y']
        df_out = df_cards_out.groupby(['field_', 'bins_', 'boundary_'])['size_','score_'].sum().reset_index()

        return df_out


    def calc_score(self, df_card, df_data: pd.DataFrame, is_woe_feature=False):
        df_score = pd.DataFrame()
        if is_woe_feature:
            pass
        else:
            bins_dict = df_card.groupby('field_').boundary_.agg(
                lambda x: sorted(set([-np.inf, np.inf] + list(x)))).to_dict()
        for field_ in df_card.field_.unique():
            if field_ in ['init_base_score','init_model_score']:
                continue
            _card = df_card[df_card.field_ == field_]
            _card["bins_"] = _card.bins_.astype(str)
            if is_woe_feature:
                _data = df_data[field_].astype(str)
            else:
                _data = pd.cut(df_data[field_].fillna(0), bins=bins_dict.get(field_, [-np.inf, np.inf])).astype(str)
            out = pd.merge(_data,_card.set_index("bins_"),how='left',left_on=_data.name,right_index=True)
            df_score[field_] = out['score_']
        df_score['init_base_score'] = df_card[df_card.field_ == 'init_base_score']['score_'].values[0]
        df_score['init_model_score'] = df_card[df_card.field_ == 'init_model_score']['score_'].values[0]
        score = df_score.sum(axis=1).apply(lambda x: round(x,2))
        return score

    def get_importance_(self):
        df_card_info = self.score_ecard.groupby("field_").agg(
            {'size_': [("weights", lambda x: list(x))],
             'score_': [("values", lambda x: list(x))]

             }).droplevel(0, 1)
        importance_ = df_card_info.apply(lambda x: get_weighted_std(x["values"], x["weights"]), axis=1)
        importance_ = (importance_/importance_.sum()).apply(lambda x:round(x,6))
        return importance_.sort_values(ascending=False)

    def predict(self, df_data):
        df_score = self.calc_score(self.score_ecard,df_data)
        return df_score

    def predict_proba(self, df_data):
        df_score = self.predict(df_data)
        df_proba =df_score.apply(lambda x:self.score_model.score_to_probability(x))
        return df_proba

    def predict_hundred(self, df_data):
        '''
        百分制
        :param df_data:
        :return:
        '''
        df_score = self.predict(df_data)
        df_hundred =df_score.apply(lambda x:round((x-self.score_lower)/(self.score_upper-self.score_lower)*100,2))
        return df_hundred

    # 单车详细评分
    def get_single_score(self, data: dict, level_threshold=None):
        score = 0
        score_detail = {}
        df_card = self.score_ecard
        for i, row in df_card.iterrows():
            field_ = row['field_']
            bins_ = row['bins_']
            score_ = row['score_']
            if field_ in ['init_base_score', 'init_model_score']:
                v = 0
            else:
                v = data.get(field_)
            try:
                v = float(v)
            except:
                score = 0
                score_detail[field_] = '异常'
                break
            if not (v > -np.inf):
                v = 0
            if v in bins_:
                score = score + score_
                score_detail[field_] = score_
            else:
                pass

        # level
        score = round(score)
        level = -1
        total_meters = data.get('run_meters', 0)
        total_seconds = data.get('run_seconds', 0)
        if (level_threshold is not None) and (total_meters >= 2000000) and (total_seconds >= 10 * 3600):
            level = np.argwhere(np.sort(level_threshold + [score]) == score).min()

        out = {'score': score, 'level': level, 'score_detail': score_detail}
        return out

    def model_explanation(self, title=None, save_path=None,**kwargs):
        return summary_explanation(self.score_ecard, title=title, save_path=save_path,**kwargs)

    def factor_report(self, title='ecard', max_display=999, save_path=None):
        return model_factor_report(self.score_ecard, title, max_display, save_path)

    def dump(self, model_name, score_bins):
        self.score_bins=score_bins
        ecard_standardized = ECardStandardized()
        ecard_standardized.dump(self, model_name, model='all')


if __name__ == '__main__':
    df_valid = pd.read_csv("data/train_test_data.csv")
    df_train_data = df_valid[df_valid['train_test_tag'] == '训练集'].fillna(0).head(1000)
    df_test_data = df_valid[df_valid['train_test_tag'] == '测试集'].fillna(0).head(1000)
    feature_columns = df_train_data.columns[4:33].tolist()
    feature_columns.extend(df_train_data.columns[36:46])
    df_X = df_train_data[feature_columns]
    df_Y = df_train_data[['label', 'label_ordinary',
                          'label_serious', 'label_major', 'label_devastating', 'label_8w','fee_got','report_fee']]
    df_Y['label']=df_Y.apply(lambda x:x['label'] if x["report_fee"]<5000 else 2,axis=1)

    ecard = ECardModel(
        kwargs_xgb={'n_estimators':2},
    )
    ecard.fit(df_X,df_Y)
    print(ecard.get_importance_())
    print(ecard.predict(df_X))
    print(ecard.predict_hundred(df_X))
    data = df_test_data.loc[0].to_dict()
    print(ecard.get_single_score(data=data, level_threshold=[-np.inf, 400, 536, 550, 600, np.inf]))
    ecard.dump('data/aa.pkl',[-np.inf, 400, 536, 550, 600, np.inf])

    import dill
    model = dill.load(open('data/aa.pkl', "rb"))
    clf_tmp = model.get("model")
    print(model.keys())
    print(clf_tmp.get_importance_())





# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2022/1/7 15:46
Desc:
'''

import numpy as np
import pandas as pd
import matplotlib as mplt
from matplotlib import pyplot as pl
import seaborn as sns
from pyecharts.charts import Bar, Line, Page, Grid
from pyecharts import options as opts

def summary_explanation(score_card, max_display=15, title=None, row_width=12, row_height=0.6, save_path=None):
    '''
    模型结构解读【评分卡构建结构 to 评分】:
        1.模型的特征构成及重要度排序(topN)，颜色越深，重要度越高；
        2.各子图宽度表示特征对模型的贡献度，对最终得分的加成，最终得分可理解为个子score_相加；
        3.各子图散点密集程度分布情况表征因子对于不同风险区间的描述精度，亦可理解为把握度高低；
    :param score_card:
    :param max_display:
    :param title:
    :param row_width:
    :param row_height:
    :param save_path:
    :return:
    '''
    # data
    df_importance = get_importance(score_card)
    size_ = min(max_display, len(df_importance))
    data = score_card.merge(pd.DataFrame(df_importance[:size_], columns=['importance_']), how='inner', left_on='field_',
                            right_index=True
                            ).sort_values('importance_', ascending=False)

    # color
    g7_color = ['#5100ad', '#5100ad', '#8d4fe8', '#a577e9', '#b490ea', '#c3a8eb', '#c9bbdd', '#dad4e3']
    g7_colormap = mplt.colors.LinearSegmentedColormap.from_list('cmap', g7_color[::-1], 256)
    g7_color_palette = sns.color_palette("Purples", n_colors=int(size_ * 1.2+10))

    # violinplot
    custom_params = {"axes.spines.left": False, "axes.spines.right": False, "axes.spines.top": False}
    sns.set_theme(style="white", rc=custom_params, palette=None)
    sns.violinplot(data=data, x="score_", y="field_", palette=g7_color_palette[::-1][10:],
                   inner="stick", orient="h", scale="width")

    # chart_init
    fig = pl.gcf().set_size_inches(row_width, size_ * row_height + 1.5)
    pl.axvline(x=0, color="#999999", zorder=-1)
    pl.suptitle(title, x=0.09, y=0.92, fontsize=16, color='#5100ad')
    pl.ylabel('')
    # pl.rcParams['font.sans-serif']=['SimHei']
    # pl.rcParams['axes.unicode_minus']=False

    # colorbar
    m = mplt.cm.ScalarMappable(cmap=g7_colormap)
    m.set_array([0, 1])
    cb = pl.colorbar(m, ticks=[0, 1], aspect=1000)
    cb.set_ticklabels(['Low', 'High'])
    cb.set_label('Feature Importance', size=12, labelpad=0)
    cb.ax.tick_params(labelsize=11, length=0)
    cb.set_alpha(1)
    cb.outline.set_visible(False)
    bbox = cb.ax.get_window_extent().transformed(pl.gcf().dpi_scale_trans.inverted())
    cb.ax.set_aspect((bbox.height - 0.9) * 20)
    if save_path is not None:
        pl.savefig(save_path, bbox_inches='tight')
        pl.close()

def sample_explanation(score_card, max_display=15, title=None, row_width=12, row_height=0.5, save_path=None):
    '''
    样本结构成分-训练集样本构成【评分卡样本构成 for 样本分布】:
        1.模型的样本分布，颜色越深，样本集中度越高，对于模型的贡献权重越高；
        2.各子图宽度表示特征对模型的贡献度，对最终得分的加成，最终得分可理解为各子score_相加；
        3.各子图密集度较高区域(颜色较深)表征该因子对于风险评估的主要作用区间，即因子作用的主区域；
    :param score_card:
    :param max_display:
    :param title:
    :param row_width:
    :param row_height:
    :param save_path:
    :return:
    '''
    # data
    df_importance = get_importance(score_card)
    size_ = min(max_display, len(df_importance))
    df_importance = df_importance[:size_]
    data = score_card.merge(pd.DataFrame(df_importance, columns=['importance_']), how='inner', left_on='field_',
                            right_index=True
                            ).sort_values('importance_', ascending=False)

    # color
    g7_color = ['#5100ad', '#5100ad', '#8d4fe8', '#a577e9', '#b490ea', '#c3a8eb', '#c9bbdd', '#dad4e3']
    g7_colormap = mplt.colors.LinearSegmentedColormap.from_list('cmap', g7_color[::-1], 256)

    # chart_init
    fig = pl.gcf().set_size_inches(row_width, size_ * row_height + 1.5)
    pl.axvline(x=0, color="#999999", zorder=-1)
    pl.suptitle(title, x=0.09, y=0.92, fontsize=16, color='#5100ad')
    pl.xlabel('score_')
    ax = pl.gca()
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')

    # scatter
    yticks_idx = []
    yticks_name = []
    feature_order = df_importance.index[::-1]
    for pos, icol_ in enumerate(feature_order):
        raw_data = data[data.field_ == icol_]
        if 'size_' in raw_data.columns:
            raw_data['size_'] = (raw_data['size_'] / raw_data['size_'].sum()).apply(lambda x: int(np.ceil(x)))
            idata = raw_data.apply(lambda x: [x['score_'] for i in range(x['size_'])], axis=1).sum()
            idata = np.array(idata)
        else:
            idata = raw_data.score_.values
        x, y = data_adapter(idata, row_height)
        x_idx = np.argsort(abs(y))
        cweight = np.array([dict(zip(x[x_idx], abs(y[x_idx]))).get(i, 1) for i in x])
        w_idx = np.argsort(cweight)
        pl.scatter(x=x[w_idx], y=pos + y[w_idx], s=20, alpha=1, cmap=g7_colormap, c=cweight[w_idx],
                   rasterized=len(x) > 10000)
        yticks_idx.append(pos)
        yticks_name.append(icol_)
    pl.yticks(yticks_idx, yticks_name, fontsize=10)

    # colorbar
    m = mplt.cm.ScalarMappable(cmap=g7_colormap)
    m.set_array([0, 1])
    cb = pl.colorbar(m, ticks=[0, 1], aspect=1000)
    cb.set_ticklabels(['Low', 'High'])
    cb.set_label('Sample Weight', size=12, labelpad=0)
    cb.ax.tick_params(labelsize=11, length=0)
    cb.set_alpha(1)
    cb.outline.set_visible(False)
    bbox = cb.ax.get_window_extent().transformed(pl.gcf().dpi_scale_trans.inverted())
    cb.ax.set_aspect((bbox.height - 0.9) * 20)

    if save_path is not None:
        pl.savefig(save_path, bbox_inches='tight')
        pl.close()


def factor_trend(score_card, max_display=15, title='因子一致性分析', subtitle='风险走势', row_width=14, row_height=14,
                 save_path=None):
    '''
    因子一致性-单调趋同【因子对于模型的功效趋势一致，对模型的单调贡献度】:
        1.因子的关联性分析，颜色越深，因子风险的趋势正向一致，颜色越浅，因子风险的趋势逆相关，颜色居中，则无关联，分布模式不同；
        2.因子对于模型的作用加成一致性
    :param score_card:
    :param max_display:
    :param title:
    :param subtitle:
    :param row_width:
    :param row_height:
    :param save_path:
    :return:
    '''
    # data
    df_importance = get_importance(score_card)
    size_ = min(max_display, len(df_importance))
    df_importance = df_importance[:size_]
    _card = score_card.merge(pd.DataFrame(df_importance, columns=['importance_']), how='inner', left_on='field_',
                             right_index=True
                             ).sort_values('importance_', ascending=False)
    _card['score_ext'] = _card.apply(lambda x: np.array([x['score_'] for i in range(x['size_'])]), axis=1)
    df_data = _card.groupby('field_').apply(
        lambda x: np.array(x["score_ext"])[np.argsort(list(x["bins_"]))]
    ).apply(lambda x: [j for i in x for j in i])
    df_data = pd.DataFrame(df_data.values.tolist(), index=df_data.index).T
    df_data.dropna(axis=1, inplace=True)
    df_corr = df_data.corr()

    # color
    g7_color_palette = sns.color_palette("binary", n_colors=size_)[::-1]
    color_dict = dict(zip(df_importance.index, g7_color_palette))
    color_list = [color_dict.get(i) for i in df_data.columns]

    # clustermap
    g = sns.clustermap(df_corr, center=0, cmap="Purples",
                       row_colors=color_list,
                       dendrogram_ratio=(.1, .15),
                       cbar_pos=(.01, .32, .03, .2),
                       linewidths=.6, figsize=(row_width, row_height),
                       )
    g.ax_row_dendrogram.remove()
    g.figure.text(0.03, 1.0, title, fontsize=14, fontfamily='SimHei', fontweight='bold')
    g.figure.text(0.03, 0.98, subtitle, fontsize=12, fontfamily='SimHei')
    ax = g.ax_heatmap
    ax.set_xlabel("")
    ax.set_ylabel("")

    if save_path is not None:
        pl.savefig(save_path, bbox_inches='tight')
        pl.close()


def single_factor_report(df_top, df_bottom, title_name='特征1风险分布', subtitle_name='tezheng1', color_list=['#5100ad']):
    '''
    :param df_top: 因子评分分布
    :param df_bottom: 因子对应样本数分布
    :param title_name:
    :param subtitle_name:
    :return:
        grid = factor_report(df_top,df_bottom)
        grid.render("tmp.html")
    '''

    def plot_base_line(df: pd.DataFrame, title_name='', subtitle_name='', xaxis_name='', yaxis_name='',
                       is_show_label=False, _interval=20):
        if isinstance(df, pd.Series):
            df = pd.DataFrame(df)
        _data = df.copy()
        _data.index = [str(i) for i in _data.index]
        c1 = Line({'bg_color': 'white'})
        c1.add_xaxis(_data.index.tolist())
        for i, icol in enumerate(_data.columns):
            c1.add_yaxis('', _data[icol].values.tolist(), is_connect_nones=True)
        c1.set_series_opts(label_opts=opts.LabelOpts(is_show=is_show_label))
        c1.set_global_opts(
            title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
            toolbox_opts=opts.ToolboxOpts(is_show=True),
            datazoom_opts=opts.DataZoomOpts(is_show=True, xaxis_index=[0, 1], range_start=0, range_end=100),
            xaxis_opts=opts.AxisOpts(name=xaxis_name),
            yaxis_opts=opts.AxisOpts(name=yaxis_name,
                                     type_="value",
                                     is_scale=True,
                                     axislabel_opts=opts.LabelOpts(formatter="{value}"),
                                     splitline_opts=opts.SplitLineOpts(is_show=True,
                                                                       linestyle_opts=opts.LineStyleOpts(opacity=0.4)),
                                     )
        )
        c1.set_colors(color_list)
        return c1

    def plot_base_bar(df: pd.DataFrame, title_name='', subtitle_name='', xaxis_name='', yaxis_name='',
                      is_show_label=None):
        if isinstance(df, pd.Series):
            df = pd.DataFrame(df)
        _data = df.copy()
        _data.index = [str(i) for i in _data.index]
        if is_show_label is None:
            is_show_label = True if len(_data) < 8 else False
        c1 = Bar({'bg_color': 'white'})
        c1.add_xaxis(_data.index.tolist())
        for i, icol in enumerate(_data.columns):
            c1.add_yaxis('', _data[icol].values.tolist())
        c1.set_series_opts(label_opts=opts.LabelOpts(is_show=is_show_label))
        c1.set_global_opts(
            title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
            toolbox_opts=opts.ToolboxOpts(is_show=True),
            xaxis_opts=opts.AxisOpts(name=xaxis_name),
            yaxis_opts=opts.AxisOpts(name=yaxis_name, type_="value", split_number=3),
            datazoom_opts=opts.DataZoomOpts(is_show=True, xaxis_index=[0, 1], range_start=0, range_end=100)
        )
        return c1

    line1 = plot_base_line(df_top, title_name=title_name, subtitle_name=subtitle_name, yaxis_name='score_')
    bar1 = plot_base_bar(df_bottom, yaxis_name='size_')
    grid = (
        Grid(init_opts=opts.InitOpts(width="800px", height="500px"))
            .add(line1, grid_opts=opts.GridOpts(pos_bottom="40%", pos_top="13%"))
            .add(bar1, grid_opts=opts.GridOpts(pos_top="70%"))
    )
    return grid

def model_factor_report(score_card, title='ecard', max_display=999, save_path=None):
    '''
    评分卡模型因子报告：【因子的评分规则详情】
        1.模型对于因子的认知程度(是否过拟合，是否欠拟合)
        2.因子表征样本的情况(样本是否有偏)
    :param score_card:
    :param title:
    :param max_display:
    :param save_path:
    :return:
    '''
    # data
    df_importance = get_importance(score_card)
    size_ = min(max_display, len(df_importance))
    df_importance = df_importance[:size_]
    page = Page(interval=5,page_title=title)
    for i, icol in enumerate(df_importance.index):
        _card = score_card[score_card.field_ == icol].sort_values('bins_').set_index('bins_')
        df_top = _card.score_.round(2)
        if 'size_' in _card.columns:
            df_bottom = _card.size_.astype(int)
        else:
            df_bottom = (df_top * 0 + 1).astype(int)
        chart_i = single_factor_report(df_top, df_bottom, title_name="{}. {}".format(i + 1, icol), subtitle_name='')
        page.add(chart_i)

    if save_path is not None:
        page.render("{}.html".format(save_path))
    else:
        return page.render_notebook()

def get_importance(score_card):
    df_card_info = score_card.groupby("field_").agg(
        {'size_': [("weights", lambda x: list(x))],
         'score_': [("values", lambda x: list(x))]

         }).droplevel(0, 1)
    get_weighted_std = lambda value_,weight_: np.average((value_-np.average(value_, weights=weight_))**2, weights=weight_)
    importance_ = df_card_info.apply(lambda x: get_weighted_std(x["values"], x["weights"]), axis=1)
    importance_ = (importance_/importance_.sum()).apply(lambda x:round(x,6))
    importance_ = importance_[importance_>0].sort_values(ascending=False)
    return importance_

def data_adapter(idata, row_height=0.5):
    inds = np.arange(len(idata))
    np.random.shuffle(inds)
    idata = idata[inds]
    N = len(idata)
    nbins = 100
    quant = np.round(nbins * (idata - np.min(idata)) / (np.max(idata) - np.min(idata) + 1e-8))
    inds = np.argsort(quant + np.random.randn(N) * 1e-6)
    layer = 0
    last_bin = -1
    ys = np.zeros(N)
    for ind in inds:
        if quant[ind] != last_bin:
            layer = 0
        ys[ind] = np.ceil(layer / 2) * ((layer % 2) * 2 - 1)
        layer += 1
        last_bin = quant[ind]
    ys *= 0.9 * (row_height / np.max(ys + 1))
    return idata, ys
